// GMEOLEModel.cpp : implementation file
//

#include "stdafx.h"
#include "gme.h"
#include "GMEOLEModel.h"
#include "GMEOLEAspect.h"
#include "GMEOLEError.h"
#include "GMEOLEColl.h"
#include "GMEDoc.h"
#include "MainFrm.h"
#include "GMEEventLogger.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


	

/////////////////////////////////////////////////////////////////////////////
// CGMEOLEModel

IMPLEMENT_DYNCREATE(CGMEOLEModel, CCmdTarget)

CGMEOLEModel::CGMEOLEModel()
{
	EnableAutomation();
}

CGMEOLEModel::~CGMEOLEModel()
{
}


void CGMEOLEModel::OnFinalRelease()
{
	// When the last reference for an automation object is released
	// OnFinalRelease is called.  The base class will automatically
	// deletes the object.  Add additional cleanup required for your
	// object before calling the base class.

	CCmdTarget::OnFinalRelease();
}

bool CGMEOLEModel::AmIValid() {
	if (CGMEDoc::theInstance) {
		POSITION pos = CGMEDoc::theInstance->GetFirstViewPosition();
		while (pos) {
			CGMEView* view = (CGMEView*)(CGMEDoc::theInstance->GetNextView(pos));
			ASSERT(view);
			if (view == m_view) {
				return true;
			}
		}
	}
	return false;
}


BEGIN_MESSAGE_MAP(CGMEOLEModel, CCmdTarget)
	//{{AFX_MSG_MAP(CGMEOLEModel)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

BEGIN_DISPATCH_MAP(CGMEOLEModel, CCmdTarget)
	//{{AFX_DISPATCH_MAP(CGMEOLEModel)
	DISP_PROPERTY_EX(CGMEOLEModel, "Active", GetActive, SetActive, VT_BOOL)
	DISP_PROPERTY_EX(CGMEOLEModel, "Aspects", GetAspects, SetAspects, VT_DISPATCH)
	DISP_PROPERTY_EX(CGMEOLEModel, "MgaModel", GetMgaModel, SetMgaModel, VT_DISPATCH)
	DISP_PROPERTY_EX(CGMEOLEModel, "Valid", GetValid, SetValid, VT_BOOL)
	DISP_FUNCTION(CGMEOLEModel, "Print", Print, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CGMEOLEModel, "PrintDialog", PrintDialog, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CGMEOLEModel, "DumpWindowsMetaFile", DumpWindowsMetaFile, VT_EMPTY, VTS_BSTR)
	DISP_FUNCTION(CGMEOLEModel, "CheckConstraints", CheckConstraints, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CGMEOLEModel, "RunComponent", RunComponent, VT_EMPTY, VTS_BSTR)
	DISP_FUNCTION(CGMEOLEModel, "RunComponentDialog", RunComponentDialog, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CGMEOLEModel, "Close", Close, VT_EMPTY, VTS_NONE)
	DISP_FUNCTION(CGMEOLEModel, "GrayOutFCO", GrayOutFCO, VT_EMPTY, VTS_BOOL VTS_BOOL VTS_DISPATCH)
	DISP_FUNCTION(CGMEOLEModel, "GrayOutAll", GrayOutAll, VT_EMPTY, VTS_BOOL)
	//}}AFX_DISPATCH_MAP
END_DISPATCH_MAP()

// Note: we add support for IID_IGMEOLEModel to support typesafe binding
//  from VBA.  This IID must match the GUID that is attached to the 
//  dispinterface in the .ODL file.

// {399A16A0-D209-4d00-9BDF-858D87EC4641}
//static const IID IID_IGMEOLEModel = 
//{ 0x399a16a0, 0xd209, 0x4d00, { 0x9b, 0xdf, 0x85, 0x8d, 0x87, 0xec, 0x46, 0x41 } };


BEGIN_INTERFACE_MAP(CGMEOLEModel, CCmdTarget)
	INTERFACE_PART(CGMEOLEModel, IID_IGMEOLEModel, Dual)
	DUAL_ERRORINFO_PART(CGMEOLEModel)
END_INTERFACE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CGMEOLEModel message handlers

void CGMEOLEModel::Print() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::Print()\r\n");

	PRECONDITION_VALID_MODEL

	m_view->SendMessage(WM_COMMAND, ID_FILE_PRINT_DIRECT);
}

void CGMEOLEModel::PrintDialog() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::PrintDialog()\r\n");

	PRECONDITION_VALID_MODEL

	m_view->SendMessage(WM_COMMAND, ID_FILE_PRINT);

}

void CGMEOLEModel::DumpWindowsMetaFile(LPCTSTR filePath) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::DumpWindowsMetaFile()\r\n");

	PRECONDITION_VALID_MODEL

	CMetaFileDC cDC;
	BOOL ret = cDC.CreateEnhanced(m_view->GetDC(),filePath,NULL,_T("GME Model"));
	if (ret == FALSE) {
		AfxMessageBox("Unable to create metafile.", MB_OK | MB_ICONSTOP);
		return;
	}

	cDC.m_bPrinting = TRUE;		// HACK by Peter (c)
	m_view->OnDraw(&cDC);

	HENHMETAFILE hEmf = cDC.CloseEnhanced();
	if ( hEmf ) {
		DeleteEnhMetaFile(hEmf);
	}
}

void CGMEOLEModel::CheckConstraints() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::CheckConstraints()\r\n");

	PRECONDITION_VALID_MODEL
	PRECONDITION_ACTIVE_CONSTMGR

	m_view->OnFileCheck();
}

BOOL CGMEOLEModel::GetActive() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GetActive()\r\n");

	PRECONDITION_VALID_MODEL

	return (CGMEView::GetActiveView() == m_view) ? TRUE : FALSE;
}

void CGMEOLEModel::SetActive(BOOL bNewValue) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::SetActive()\r\n");

	PRECONDITION_VALID_MODEL;

	if (bNewValue == TRUE) {
		CMainFrame::theInstance->ActivateView(m_view);
	}
}

void CGMEOLEModel::RunComponent(LPCTSTR appID) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::RunComponent()\r\n");

	PRECONDITION_VALID_MODEL

	m_view->RunComponent(appID);
}

void CGMEOLEModel::RunComponentDialog() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::RunComponentDialog()\r\n");

	PRECONDITION_VALID_MODEL

	m_view->RunComponent("");
}

LPDISPATCH CGMEOLEModel::GetAspects() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GetAspects()\r\n");

	PRECONDITION_VALID_MODEL

	CGMEOLEColl* coll = new CGMEOLEColl();

	POSITION pos = m_view->guiMeta->aspects.GetHeadPosition();
	while (pos) {
		CGuiMetaAspect* guiaspect = (CGuiMetaAspect*)(m_view->guiMeta->aspects.GetNext(pos));
		ASSERT(guiaspect);
		CGMEOLEAspect* aspect = new CGMEOLEAspect();
		aspect->SetGuiParams(m_view, guiaspect);
		coll->Add(aspect->GetIDispatch(FALSE));
		aspect->GetIDispatch(FALSE)->Release();	// We do not hold reference, hopefully the collection will hold one
	}

	return coll->GetIDispatch(FALSE);
}

void CGMEOLEModel::SetAspects(LPDISPATCH) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::SetAspects()\r\n");

	SetNotSupported();
}

LPDISPATCH CGMEOLEModel::GetMgaModel() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GetMgaModel()\r\n");

	PRECONDITION_VALID_MODEL

	return m_view->currentModel;

	return NULL;
}

void CGMEOLEModel::SetMgaModel(LPDISPATCH) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::SetMgaModel()\r\n");

	SetNotSupported();
}

BOOL CGMEOLEModel::GetValid() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GetValid()\r\n");
	
	// PRECONDITIONS: None

	return (AmIValid() ? TRUE : FALSE);
}

void CGMEOLEModel::SetValid(BOOL) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::SetValid()\r\n");

	SetNotSupported();
}

void CGMEOLEModel::Close() 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::Close()\r\n");

	PRECONDITION_VALID_MODEL;

	m_view->GetParentFrame()->DestroyWindow();
}

void CGMEOLEModel::GrayOutFCO(BOOL bGray, BOOL bNeighbours, LPDISPATCH mgaFCO) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GrayOutFCO()\r\n");

	PRECONDITION_VALID_MODEL;

	CComPtr<IDispatch> alienFCO(mgaFCO);
	CComPtr<IMgaFCO> mFCO;
	if (!SUCCEEDED(alienFCO.QueryInterface(&mFCO))) {
		return;
	}

	// mgaFCO is comming from an unknown territory, do not use it as-is. (except for this search)
	CGuiObject *gObj = CGuiObject::FindObject(mFCO, m_view->children);
	CGuiConnection *gConn = CGuiConnection::FindConnection(mFCO, m_view->connections);

	if (gObj) {
		gObj->GrayOut(bGray == TRUE);
		if (bNeighbours == TRUE) {
			gObj->GrayOutNeighbors();
		}
		CGuiFco::GrayOutNonInternalConnections(m_view->connections);
	}
	else if (gConn) {
		gConn->GrayOut(bGray == TRUE);
		gConn->GrayOutEndPoints();
		if (bNeighbours == TRUE) {
			CGuiFco::GrayOutNonInternalConnections(m_view->connections);
		}
	}
	m_view->Invalidate();

}

void CGMEOLEModel::GrayOutAll(BOOL bGray) 
{
	CGMEEventLogger::LogGMEEvent("CGMEOLEModel::GrayOutAll()\r\n");

	PRECONDITION_VALID_MODEL;

	CGuiFco::GrayOutFcos(m_view->children, (bGray == TRUE));
	CGuiFco::GrayOutFcos(m_view->connections, (bGray == TRUE));
	m_view->Invalidate();
}

DELEGATE_DUAL_INTERFACE(CGMEOLEModel, Dual)

// Implement ISupportErrorInfo to indicate we support the
// OLE Automation error handler.
IMPLEMENT_DUAL_ERRORINFO(CGMEOLEModel, IID_IGMEOLEModel)


STDMETHODIMP CGMEOLEModel::XDual::put_Active(VARIANT_BOOL isActive)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->SetActive(isActive);
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::get_Active(VARIANT_BOOL* isActive)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
#pragma warning(disable: 4310) // cast truncates constant value
		*isActive = (pThis->GetActive() == FALSE) ? VARIANT_FALSE : VARIANT_TRUE;
#pragma warning(default: 4310) // cast truncates constant value
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::get_Aspects(IGMEOLEColl** coll)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		LPDISPATCH lpDisp = pThis->GetAspects();
		lpDisp->QueryInterface(IID_IGMEOLEColl, (LPVOID*)coll);
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::get_MgaModel(IMgaModel** model)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		LPDISPATCH lpDisp = pThis->GetMgaModel();
		lpDisp->QueryInterface(IID_IMgaModel, (LPVOID*)model);
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::get_Valid(VARIANT_BOOL* isValid)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
#pragma warning(disable: 4310) // cast truncates constant value
		*isValid = (pThis->GetValid() == FALSE) ? VARIANT_FALSE : VARIANT_TRUE;
#pragma warning(default: 4310) // cast truncates constant value
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::Print()
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->Print();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}

STDMETHODIMP CGMEOLEModel::XDual::PrintDialog()
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->PrintDialog();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::DumpWindowsMetaFile(BSTR filePath)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->DumpWindowsMetaFile(CString(filePath));
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::CheckConstraints()
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->CheckConstraints();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::RunComponent(BSTR appID)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->RunComponent(CString(appID));
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::RunComponentDialog()
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->RunComponentDialog();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::Close()
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->Close();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::GrayOutFCO(VARIANT_BOOL bGray, VARIANT_BOOL bNeighbours, IMgaFCOs* mgaFCO)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		LPDISPATCH lpDisp = NULL;
		mgaFCO->QueryInterface(IID_IDispatch, (LPVOID*)&lpDisp);
		pThis->GrayOutFCO(bGray, bNeighbours, lpDisp);
		lpDisp->Release();
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


STDMETHODIMP CGMEOLEModel::XDual::GrayOutAll(VARIANT_BOOL bGray)
{
	METHOD_PROLOGUE(CGMEOLEModel, Dual)

	TRY_DUAL(IID_IGMEOLEModel)
	{
		pThis->GrayOutAll(bGray);
		return NOERROR;
	}
	CATCH_ALL_DUAL
}


