// MgaComplexOps.cpp : Implementation of FCO Copy/Move/Derive/Delete operations
#include "stdafx.h"
#include "MgaFCO.h"


/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// DELETE //////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////


void SingleObjTreeDelete(CoreObj &self, bool deleteself = true) {
	CComPtr<ICoreAttributes> atts;
	COMTHROW(self->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
		attrid_type ai;
		CComPtr<ICoreMetaAttribute> mattr;
		COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
		COMTHROW(mattr->get_AttrID(&ai));
		if(ai >= ATTRID_COLLECTION) {
			ai -= ATTRID_COLLECTION;
			if(LINKREF_ATTR(ai) && ai != ATTRID_PARENT) {
				CoreObjs collmembers = self[ai + ATTRID_COLLECTION];
				ITERATE_THROUGH(collmembers) {
					SingleObjTreeDelete(ITER);
				}
			}
		}
	} MGACOLL_ITERATE_END;
	if(deleteself) COMTHROW(self->Delete());
}


void FCO::inDeleteObject() { 
			long status;
			COMTHROW(get_Status(&status));
			metaid_type typ = GetMetaID(self);
			if(status == OBJECT_DELETED) return;  // since collections contain objects, it may happen!!!
// Step 1: delete structure of this object, but not the object itself
			if(typ == DTID_CONNECTION) {
				ITERATE_THROUGH(self[ATTRID_CONNROLE+ATTRID_COLLECTION]) CoreObjMark(ITER[ATTRID_XREF], OBJEVENT_DISCONNECTED);
			}
			else if(typ == DTID_SET) {
				ITERATE_THROUGH(self[ATTRID_SETMEMBER+ATTRID_COLLECTION]) CoreObjMark(ITER[ATTRID_XREF], OBJEVENT_SETEXCLUDED);
			}
			else if(typ == DTID_REFERENCE) {
				CoreObj h = self[ATTRID_REFERENCE]; if(h) CoreObjMark(h, OBJEVENT_REFRELEASED);
			}

			if(typ != DTID_FOLDER) {
				CoreObjs children = self[ATTRID_DERIVED + ATTRID_COLLECTION];
				ITERATE_THROUGH(children) ObjForCore(ITER)->inDeleteObject();
			}

			SingleObjTreeDelete(self, false);

// Step 2: delete children, and derivated objects
			if(typ == DTID_MODEL || typ == DTID_FOLDER) {
				CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
				ITERATE_THROUGH(children) ObjForCore(ITER)->inDeleteObject();
			}
// Step 3: mark or delete objects that are affected
			CoreObj parent;   
			if(typ != DTID_FOLDER) {
				parent = self[ATTRID_FCOPARENT];
// REFERENCES	set references to empty
				CoreObj const nil;  // Null COM pointer
				{
					CoreObjs referringobjs = self[ATTRID_REFERENCE + ATTRID_COLLECTION];
					ITERATE_THROUGH(referringobjs) {
						if(ITER.IsDeleted()) continue;   // object was deleted during a previous iteration of the same cycle
						int todo = MODEMASK(MM_REF, MM_INTO);
						if(todo != MM_ERROR && MODEFLAG(MM_REF,MM_FULLDELETE)) {
							ObjForCore(ITER)->inDeleteObject();
						}
						else 	{
							setcheck(mgaproject, ITER, todo == MM_ERROR ? CHK_ILLEGAL : CHK_CHANGED);
							ITER[ATTRID_REFERENCE] = nil;
							CoreObjMark(ITER, OBJEVENT_RELATION);
						}
					}
				}
// OTHER types of xrefs:
				CoreObjs xrefs = self[ATTRID_XREF + ATTRID_COLLECTION];
				ITERATE_THROUGH(xrefs) {
					if(ITER.IsDeleted()) continue; // object was deleted during a previous iteration of the same cycle
					switch(GetMetaID(ITER)) {
					case DTID_CONNROLE:
// Connection role
						{
						CoreObj fco = ITER.GetMgaObj();
						int todo = MODEMASK(MM_CONN, MM_INTO);
						if(todo != MM_ERROR && MODEFLAG(MM_CONN,MM_FULLDELETE)) {
							ObjForCore(fco)->inDeleteObject();
						}
						else 	{
							setcheck(mgaproject, fco, todo == MM_ERROR ? CHK_ILLEGAL : CHK_CHANGED);
							SingleObjTreeDelete(ITER);
							CoreObjMark(fco, OBJEVENT_RELATION);
						}
						break;
						}
					case DTID_SETNODE:
// Set
						{
						CoreObj fco = ITER.GetMgaObj();
						int todo = MODEMASK(MM_SET, MM_INTO);
						if(todo != MM_ERROR && MODEFLAG(MM_SET,MM_FULLDELETE)) {
							ObjForCore(fco)->inDeleteObject();
						}
						else 	{
							setcheck(mgaproject, fco, todo == MM_ERROR ? CHK_ILLEGAL : CHK_CHANGED);
							SingleObjTreeDelete(ITER);
							CoreObjMark(fco, OBJEVENT_RELATION);
						}
						break;
						}
					case DTID_REFATTR:
// REF attribute:  set it to NIL
						ITER[ATTRID_XREF] = nil;
						CoreObjMark(ITER[ATTRID_ATTRPARENT], OBJEVENT_ATTR);
						break;
					case DTID_REGNODE:
// registry REF attribute:  set it to NIL
						ITER[ATTRID_XREF] = nil;
						CoreObjMark(ITER[ATTRID_REGNOWNER], OBJEVENT_REGISTRY);
						break;
					default:	
										COMTHROW(E_MGA_META_INCOMPATIBILITY);
					}
				}
				if(typ == DTID_REFERENCE) {
					CoreObjs segrefs = self[ATTRID_SEGREF + ATTRID_COLLECTION];
					ITERATE_THROUGH(segrefs) {
					    if(ITER.IsDeleted()) continue; // object was deleted during a previous iteration of the same cycle
						CoreObj fco = ITER.GetMgaObj();
						int todo = MODEMASK(MM_CONN, MM_INTO);
						if(todo != MM_ERROR && MODEFLAG(MM_CONN,MM_FULLDELETE)) {
							ObjForCore(fco)->inDeleteObject();
						}
						else 	{
							setcheck(mgaproject, fco, todo == MM_ERROR ? CHK_ILLEGAL : CHK_CHANGED);
							SingleObjTreeDelete(CoreObj(ITER[ATTRID_CONNSEG]));
							CoreObjMark(fco, OBJEVENT_RELATION);
						}
					}
				}
			} 
			else parent = self[ATTRID_FPARENT];

//Step4: delete the object itself
			COMTHROW(self->Delete());
			SelfMark(OBJEVENT_DESTROYED);
			CoreObjMark(parent, OBJEVENT_LOSTCHILD);
}

HRESULT FCO::DeleteObject() { 
		COMTRY_IN_TRANSACTION {
			CheckWrite();

			if(self[ATTRID_PERMISSIONS] & LIBRARY_FLAG) COMTHROW(E_MGA_OP_REFUSED);
			// check for non-primary deerived
			if(self[ATTRID_RELID] >= RELIDSPACE) COMTHROW(E_MGA_OP_REFUSED);
			// check for rootfolder
			if(!CoreObj(self[ATTRID_PARENT]).IsContainer()) {  
				COMTHROW(E_MGA_OP_REFUSED);  
			}
			inDeleteObject();

			docheck(mgaproject);

		} COMCATCH_IN_TRANSACTION(;);
}


/////////////////////////////////////////////////////////////////////////////////
///////////////////////// COPY, MOVE, DERIVE ////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

// take second objects from list1, and put them to the first place into list2
void shiftlist(coreobjpairhash &list1, coreobjhash &list2) {
		coreobjpairhash::iterator ii, beg = list1.begin(), end = list1.end();
		for(ii = beg; ii != end; ii++) list2.insert(coreobjhash::value_type((*ii).second, 0));
}


void ObjTreeCollect(CMgaProject *mgaproject, CoreObj &self, coreobjhash &crealist, int code ) {
	metaid_type s = s = GetMetaID(self);
	if(s >= DTID_MODEL && s <= DTID_SET) {
		crealist.insert(coreobjhash::value_type(self, 0));
		setcheck(mgaproject, self, code);
	}
	if(s == DTID_MODEL) {
			CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
			ITERATE_THROUGH(children) {
				ObjTreeCollect(mgaproject, ITER, crealist, code);
			}
	}
}

// by ZolMol -invented by
void ObjTreeCollectFoldersToo(CMgaProject *mgaproject, CoreObj &self, coreobjhash &crealist, int code ) {
	metaid_type s = s = GetMetaID(self);
	if(s >= DTID_MODEL && s <= DTID_FOLDER) {
		crealist.insert(coreobjhash::value_type(self, 0));
		setcheck(mgaproject, self, code);
	}
	if(s == DTID_MODEL) {
			CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
			ITERATE_THROUGH(children) {
				ObjTreeCollectFoldersToo(mgaproject, ITER, crealist, code);
			}
	}
	if(s == DTID_FOLDER) {
			CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
			ITERATE_THROUGH(children) {
				ObjTreeCollectFoldersToo(mgaproject, ITER, crealist, code);
			}
	}
}

void ObjTreeCopy(CMgaProject *mgaproject, CoreObj self, CoreObj &nobj, coreobjpairhash &crealist)
{
	metaid_type s;
	COMTHROW(mgaproject->dataproject->CreateObject(s = GetMetaID(self), &nobj.ComPtr()));
	if(s >= DTID_MODEL && s <= DTID_SET) {
		crealist.insert(coreobjpairhash::value_type(self, nobj));
		setcheck(mgaproject, nobj, CHK_NEW);
		CoreObjMark(nobj, OBJEVENT_CREATED);
	}

	CComPtr<ICoreAttributes> atts;
	COMTHROW(self->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
			attrid_type ai;
			CComPtr<ICoreMetaAttribute> mattr;
			COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
			COMTHROW(mattr->get_AttrID(&ai));
			if(ai < ATTRID_COLLECTION) {
				// remove library flags from a copy
				if(ai == ATTRID_PERMISSIONS) nobj[ai] = self[ai] & INSTANCE_FLAG;
				else nobj[ai] = static_cast<CComVariant>(self[ai]);
			}
			else {
				ai -= ATTRID_COLLECTION;
				if(LINKREF_ATTR(ai)) {
					CoreObjs collmembers = self[ai + ATTRID_COLLECTION];
					ITERATE_THROUGH(collmembers) {
						CoreObj nchild;
						ObjTreeCopy(mgaproject, ITER, nchild, crealist);
						nchild[ai] = nobj;
					}
				}
			}

	} MGACOLL_ITERATE_END;
}


void ObjTreeCopyFoldersToo(CMgaProject *mgaproject, CoreObj self, CoreObj &nobj, coreobjpairhash &crealist) {
	metaid_type s;
	COMTHROW(mgaproject->dataproject->CreateObject(s = GetMetaID(self), &nobj.ComPtr()));
	if(s >= DTID_MODEL && s <= DTID_FOLDER) {
		crealist.insert(coreobjpairhash::value_type(self, nobj));
		setcheck(mgaproject, nobj, CHK_NEW);
		CoreObjMark(nobj, OBJEVENT_CREATED);
	}

	CComPtr<ICoreAttributes> atts;
	COMTHROW(self->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
			attrid_type ai;
			CComPtr<ICoreMetaAttribute> mattr;
			COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
			COMTHROW(mattr->get_AttrID(&ai));
			if(ai < ATTRID_COLLECTION) {
				// remove library flags from a copy
				if(ai == ATTRID_PERMISSIONS) nobj[ai] = self[ai] & INSTANCE_FLAG;
				else nobj[ai] = static_cast<CComVariant>(self[ai]);
			}
			else {
				ai -= ATTRID_COLLECTION;
				if(LINKREF_ATTR(ai)) {
					CoreObjs collmembers = self[ai + ATTRID_COLLECTION];
					ITERATE_THROUGH(collmembers) {
						CoreObj nchild;
						ObjTreeCopy(mgaproject, ITER, nchild, crealist);
						nchild[ai] = nobj;
					}
				}
			}

	} MGACOLL_ITERATE_END;
}

// get the derived-chain distance of the closest real basetype (or if there is none, the length of the full base chain) 
// if the object is not derived, or it is a real subtype/instance (i.e derived not because of its parent) 0 is returned

int GetRealSubtypeDist(CoreObj oldobj) {
			CoreObj pp = oldobj;
			int derdist = 0;
			while(pp[ATTRID_RELID] >= RELIDSPACE) { 
				pp = pp[ATTRID_DERIVED];
				ASSERT(pp);
				derdist++;
			}
			return derdist;
}

void ObjTreeDist(CoreObj self, int derdist) {
	CComPtr<ICoreAttributes> atts;
	COMTHROW(self->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
			attrid_type ai;
			CComPtr<ICoreMetaAttribute> mattr;
			COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
			COMTHROW(mattr->get_AttrID(&ai));
			if(ai == ATTRID_DERIVED) {
				CoreObj h = self;
				for(int i = 0; i < derdist && (h = h[ATTRID_DERIVED]); i++) {
					MergeRegs(h,self);
					MergeAttrs(h,self);
				}
				CoreObj p;
				self[ai] = p = self.FollowChain(ai, derdist+1);
				if(p) self[ATTRID_PERMISSIONS] = p[ATTRID_PERMISSIONS] & INSTANCE_FLAG;
				else self[ATTRID_PERMISSIONS] = 0;
			}
			else if(ai == ATTRID_MASTEROBJ) {
					self[ai] = self.FollowChain(ai, derdist+1);
			}
			if(ai == ATTRID_FCOPARENT + ATTRID_COLLECTION ||
			   ai == ATTRID_CONNROLE + ATTRID_COLLECTION ||
			   ai == ATTRID_SETMEMBER + ATTRID_COLLECTION) {
				CoreObjs collmembers = self[ai];
				ITERATE_THROUGH(collmembers) {
							ObjTreeDist(ITER, derdist);
				}
			}

	} MGACOLL_ITERATE_END;
}	


/////////////////
// ObjTreeDerive
/////////////////
// Create a derived deep copy of 'origobj' to 'newobj'
// -- 'newobj' tree is exact copy of 'origobj' tree, except 
//			ATTRID_INSTANCE: set if instance, copied from 'origobj' tree othervise
//			ATTRID_DERIVED & ATTRID_MASTEROBJ: point to the corresponding object in 'origobj' tree
// -- on return the new root object 'newobj' will also be contained in the container of 'origobj'.
// -- new FCO-s are inserted in 'crealist'
///////////////////
// !!! RECURSIVE 
///////////////////
void ObjTreeDerive(CMgaProject *mgaproject, const CoreObj &origobj, CoreObj &newobj, coreobjpairhash &crealist, long instance) {
	metaid_type s;
	COMTHROW(mgaproject->dataproject->CreateObject(s = origobj.GetMetaID(), &newobj.ComPtr()));
	if(s >= DTID_MODEL && s <= DTID_SET) {
		crealist.insert(coreobjpairhash::value_type(origobj, newobj));
		setcheck(mgaproject, newobj, CHK_NEW);
		CoreObjMark(newobj, OBJEVENT_CREATED);
	}
	
	CComPtr<ICoreAttributes> atts;
	COMTHROW(origobj->get_Attributes(&atts));
	MGACOLL_ITERATE(ICoreAttribute, atts) {
			attrid_type ai;
			CComPtr<ICoreMetaAttribute> mattr;
			COMTHROW(MGACOLL_ITER->get_MetaAttribute(&mattr));
			COMTHROW(mattr->get_AttrID(&ai));
			if(ai < ATTRID_COLLECTION) {
				switch(ai) {
				case ATTRID_DERIVED: 
				case ATTRID_MASTEROBJ:
					newobj[ai] = origobj;
					break;
				case ATTRID_PERMISSIONS:
						// overwrite copy only if instance
					// PETER'S FIX 
					// newobj[ai] = (origobj[ai] | long(instance)) & INSTANCE_FLAG;
					newobj[ai] = (origobj[ai] | long(instance)) & INSTANCE_FLAG;
					// PETER END
					break;
				case ATTRID_RELID: 
					{
					long relid = origobj[ai]+RELIDSPACE;
					if(relid <= 0) COMTHROW(E_MGA_LONG_DERIVCHAIN);  // overflow
					newobj[ai] = relid;
					}
					break;
				case ATTRID_LASTRELID: 
					newobj[ai] = 0L;
					break;
				default:
					newobj[ai] = static_cast<CComVariant>(origobj[ai]);
				}
			}
			else {
				ai -= ATTRID_COLLECTION;
				if(LINKREF_ATTR(ai) && ai != ATTRID_ATTRPARENT && ai != ATTRID_REGNOWNER) {
					CoreObjs collmembers = origobj[ai + ATTRID_COLLECTION];
					ITERATE_THROUGH(collmembers) {
						CoreObj nchild;
						ObjTreeDerive(mgaproject, ITER, nchild, crealist, instance);
						nchild[ai] = newobj;
					}
					if(ai == ATTRID_FCOPARENT && collmembers.Count() > 0) {
							CoreObjMark(newobj, OBJEVENT_NEWCHILD);
					}
				}
			}

	} MGACOLL_ITERATE_END;
}



/////////////////
// ObjTreeReconnect
/////////////////
// Reconnects relations in a newly copied/derived tree
// 
// -- 'crealist'contains old-new equivalent pairs for quick access
// -- if object is not in crealist, it may still be reconnected 
//		if it is an internal reference, (i.e., the ends of the pointer belong to the same rootFCO).
//	  -- if the relation has a masterobj, then the relation of the root masterobj is checked
//	  -- otherwise the target of the ralation must be contained in the crealist (list of newly created objects)
// -- return value used internally to indicate that the tree contained relations which were not reconnected
///////////////////
// !!! RECURSIVE 
///////////////////
bool ObjTreeReconnect(CoreObj self, coreobjpairhash &crealist, CoreObj const &derivtgt) {
	typedef struct { metaid_type mid; int search, reconnect; } table;
	static const table tab[] = {
		{DTID_MODEL,	ATTRID_FCOPARENT,	0},
		{DTID_REFERENCE, 0,					ATTRID_REFERENCE},
		{DTID_SET,		ATTRID_SETMEMBER,	0},
		{DTID_SETNODE,	0,					ATTRID_XREF },
		{DTID_CONNECTION, ATTRID_CONNROLE,	0},
		{DTID_CONNROLE, ATTRID_CONNSEG,		ATTRID_XREF},
		{DTID_CONNROLESEG, 0,				ATTRID_SEGREF}
	};


	bool containedexternal = false;
	metaid_type n = GetMetaID(self);

	for(int i = 0; i < sizeof(tab) / sizeof(tab[0]); i++) {
		if(tab[i].mid == n) {
			const table *ll = &tab[i];
			if(ll->search) {
				CoreObjs children = self[ll->search+ATTRID_COLLECTION];
				if(n == DTID_CONNROLE) {
					children.Sort();
				}
				ITERATE_THROUGH(children) {
					if(ObjTreeReconnect(ITER,crealist, derivtgt)) {
						containedexternal = true;
						if(n == DTID_CONNROLE) break; // stop processing rolesegments after an ext ref.
					}
				}
			}
			if(!containedexternal && ll->reconnect) {
			  // the relationship pointer will to be reconnected if
				// - the object has no master (only possible with copied objects)
				// - or the relation in the most original base is internal
			  if(!self.GetMaster() || IsInternalRelation(self)) {  
				    CoreObj oldr = self[ll->reconnect];
					coreobjpairhash::iterator i = crealist.find(oldr);
					if(i != crealist.end()) {  
// if it is a reference to a newly created object: use lookup
						self[ll->reconnect] = (*i).second;
					}
					else if(derivtgt) {  
						ASSERT(CoreObj(derivtgt[ATTRID_DERIVED]));
						ASSERT(derivtgt[ATTRID_RELID] < RELIDSPACE);
						CoreObj newr;
						GetDerivedEquivalent(oldr, derivtgt, newr);
						self[ll->reconnect] = newr;
					}
			  }
			  else containedexternal = true;
			}
			break;
		}
	}
	return containedexternal;
}

bool ObjTreeReconnectFoldersToo(CoreObj self, coreobjpairhash &crealist, CoreObj const &derivtgt) {
	typedef struct { metaid_type mid; int search, reconnect; } table;
	static const table tab[] = {
		{DTID_FOLDER,	ATTRID_FCOPARENT,	0},
		{DTID_MODEL,	ATTRID_FCOPARENT,	0},
		{DTID_REFERENCE, 0,					ATTRID_REFERENCE},
		{DTID_SET,		ATTRID_SETMEMBER,	0},
		{DTID_SETNODE,	0,					ATTRID_XREF },
		{DTID_CONNECTION, ATTRID_CONNROLE,	0},
		{DTID_CONNROLE, ATTRID_CONNSEG,		ATTRID_XREF},
		{DTID_CONNROLESEG, 0,				ATTRID_SEGREF}
	};


	bool containedexternal = false;
	metaid_type n = GetMetaID(self);

	for(int i = 0; i < sizeof(tab) / sizeof(tab[0]); i++) {
		if(tab[i].mid == n) {
			const table *ll = &tab[i];
			if(ll->search) {
				CoreObjs children = self[ll->search+ATTRID_COLLECTION];
				if(n == DTID_CONNROLE) {
					children.Sort();
				}
				ITERATE_THROUGH(children) {
					if(ObjTreeReconnect(ITER,crealist, derivtgt)) {
						containedexternal = true;
						if(n == DTID_CONNROLE) break; // stop processing rolesegments after an ext ref.
					}
				}
			}
			if(!containedexternal && ll->reconnect) {
			  // the relationship pointer will to be reconnected if
				// - the object has no master (only possible with copied objects)
				// - or the relation in the most original base is internal
			  if(!self.GetMaster() || IsInternalRelation(self)) {  
				    CoreObj oldr = self[ll->reconnect];
					coreobjpairhash::iterator i = crealist.find(oldr);
					if(i != crealist.end()) {  
// if it is a reference to a newly created object: use lookup
						self[ll->reconnect] = (*i).second;
					}
					else if(derivtgt) {  
						ASSERT(CoreObj(derivtgt[ATTRID_DERIVED]));
						ASSERT(derivtgt[ATTRID_RELID] < RELIDSPACE);
						CoreObj newr;
						GetDerivedEquivalent(oldr, derivtgt, newr);
						self[ll->reconnect] = newr;
					}
			  }
			  else containedexternal = true;
			}
			break;
		}
	}
	return containedexternal;
}



/////////////////
// ObjTreeCheckRelations
/////////////////
// Removes outgoing/internal relations (or whole objects) in a newly copied/derived/moved tree
// The policy is based on the project-wide operationsmask queried with MM_OUTOF or MM_INTERNAL masks
// 
// -- 'self' is the root of the tree, 
// -- 'internals' are all the objects that are considered internal for this operation
// -- all relations hold by references, sets and connections are checked
// -- if the required operation is
//	- MM_ERROR: error is signaled
//  - MM_CLEAR: the corresponding pointer is cleared
//  - MM_CLEAR & MM_FULLDELETE: the object itself is cleared
//  - MM_DO:    nothing happens 
///////////////////
// !!! RECURSIVE 
///////////////////
void ObjTreeCheckRelations(CMgaProject *mgaproject, CoreObj &self, coreobjhash &internals) {
	typedef struct { metaid_type mid; int search, check, mm; } table;
	static const table tab[] = {
		{DTID_MODEL, ATTRID_FCOPARENT, 0},
		{DTID_REFERENCE, 0, ATTRID_REFERENCE, MM_REF},
		{DTID_SET, ATTRID_SETMEMBER, 0},
		{DTID_SETNODE, 0, ATTRID_XREF, MM_SET },
		{DTID_CONNECTION, ATTRID_CONNROLE, 0},
		{DTID_CONNROLE, ATTRID_CONNSEG, ATTRID_XREF, MM_CONN},
		{DTID_CONNROLESEG, 0, ATTRID_SEGREF}
	};

	metaid_type n = GetMetaID(self);

	for(int i = 0; i < sizeof(tab) / sizeof(tab[0]); i++) {
		if(tab[i].mid == n) {
			const table *ll = &tab[i];
			if(ll->search) {
				CoreObjs children = self[ll->search+ATTRID_COLLECTION];
				ITERATE_THROUGH(children) {
					if(ITER.IsDeleted()) continue;
					ObjTreeCheckRelations(mgaproject, ITER, internals);
				}
			}
			if(ll->check && !self.IsDeleted()) {
				CoreObj tgt = self[ll->check];
				int mm2 = (internals.find(tgt) != internals.end()) ? MM_INTERNAL : MM_OUTOF;
				switch(MODEMASK(ll->mm, mm2)) {
				case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
				case MM_CLEAR: 
					self[ll->check] = NULLCOREOBJ;
					CoreObj selfobj = self.GetMgaObj();
					if( (GetMetaID(selfobj) == DTID_CONNECTION && ObjForCore(selfobj)->simpleconn()) ||
						MODEFLAG(ll->mm, MM_FULLDELETE)) {
							ObjForCore(selfobj)->inDeleteObject();
					}
					else {
						if(n != DTID_REFERENCE) {
							if(n == DTID_CONNROLESEG) self = self[ATTRID_CONNSEG];
							SingleObjTreeDelete(self);
						}
					}
				}
			}
			break;
		}
	}
}

void ObjTreeCheckRelationsFoldersToo(CMgaProject *mgaproject, CoreObj &self, coreobjhash &internals) {
	typedef struct { metaid_type mid; int search, check, mm; } table;
	static const table tab[] = {
		{DTID_FOLDER, ATTRID_FCOPARENT, 0},
		{DTID_MODEL, ATTRID_FCOPARENT, 0},
		{DTID_REFERENCE, 0, ATTRID_REFERENCE, MM_REF},
		{DTID_SET, ATTRID_SETMEMBER, 0},
		{DTID_SETNODE, 0, ATTRID_XREF, MM_SET },
		{DTID_CONNECTION, ATTRID_CONNROLE, 0},
		{DTID_CONNROLE, ATTRID_CONNSEG, ATTRID_XREF, MM_CONN},
		{DTID_CONNROLESEG, 0, ATTRID_SEGREF}
	};

	metaid_type n = GetMetaID(self);

	for(int i = 0; i < sizeof(tab) / sizeof(tab[0]); i++) {
		if(tab[i].mid == n) {
			const table *ll = &tab[i];
			if(ll->search) {
				CoreObjs children = self[ll->search+ATTRID_COLLECTION];
				ITERATE_THROUGH(children) {
					if(ITER.IsDeleted()) continue;
					ObjTreeCheckRelationsFoldersToo(mgaproject, ITER, internals);
				}
			}
			if(ll->check && !self.IsDeleted()) {
				CoreObj tgt = self[ll->check];
				int mm2 = (internals.find(tgt) != internals.end()) ? MM_INTERNAL : MM_OUTOF;
				switch(MODEMASK(ll->mm, mm2)) {
				case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
				case MM_CLEAR: 
					self[ll->check] = NULLCOREOBJ;
					CoreObj selfobj = self.GetMgaObj();
					if( (GetMetaID(selfobj) == DTID_CONNECTION && ObjForCore(selfobj)->simpleconn()) ||
						MODEFLAG(ll->mm, MM_FULLDELETE)) {
							ObjForCore(selfobj)->inDeleteObject();
					}
					else {
						if(n != DTID_REFERENCE) {
							if(n == DTID_CONNROLESEG) self = self[ATTRID_CONNSEG];
							SingleObjTreeDelete(self);
						}
					}
				}
			}
			break;
		}
	}
}

/////////////////
// ObjTreeCheckINTORelations
/////////////////
// Removes incoming relations (or whole objects) that point into a newly copied/derived/moved tree
// The policy is based on the project-wide operationsmask queried with MM_INTO mask
// 
// -- 'self' is the root of the tree, 
// -- 'internals' are all the objects that are considered internal for this operation
// -- all relations that point to objects in the tree are checked.
// -- if the required operation is
//	- MM_ERROR: error is signaled
//  - MM_CLEAR: the corresponding pointer is cleared
//  - MM_CLEAR & MM_FULLDELETE: the object itself is cleared
//  - MM_DO:    nothing happens 
///////////////////
// !!! RECURSIVE 
///////////////////
void ObjTreeCheckINTORelations(CMgaProject *mgaproject, CoreObj &self, coreobjhash &internals) {
	metaid_type n = GetMetaID(self);
	ASSERT(n >= DTID_MODEL && n <= DTID_SET);
	CoreObjs xrefs = self[ATTRID_XREF + ATTRID_COLLECTION]; 
	ITERATE_THROUGH(xrefs) {
		metaid_type st = GetMetaID(ITER);
		if(st == DTID_SETNODE || st == DTID_CONNROLE) {
			CoreObj rel_owner = ITER.GetMgaObj();
			if(internals.find(rel_owner) == internals.end()) {
				int ttt = st == DTID_CONNROLE ? MM_CONN : MM_SET;
				setcheck(mgaproject, rel_owner, CHK_CHANGED);
				switch(MODEMASK(ttt, MM_INTO)) {
				case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
					break;
				case MM_CLEAR: 
					if( st == DTID_CONNROLE && ObjForCore(rel_owner)->simpleconn() ||
					    MODEFLAG(ttt, MM_FULLDELETE)) {
					    ObjForCore(rel_owner)->inDeleteObject();
					} 
					else {
						CoreObjMark(self, st == DTID_CONNROLE ? OBJEVENT_DISCONNECTED : OBJEVENT_SETEXCLUDED);
						CoreObjMark(rel_owner, OBJEVENT_RELATION);
						SingleObjTreeDelete(ITER);
					}
					break;
				}
			}	
		}
	}
	{
	CoreObjs refs = self[ATTRID_REFERENCE + ATTRID_COLLECTION]; 
	ITERATE_THROUGH(refs) {
		CoreObj rel_owner = ITER;
		if(internals.find(rel_owner) == internals.end()) {
			setcheck(mgaproject, rel_owner, CHK_CHANGED);
			switch(MODEMASK(MM_REF, MM_INTO)) {
			case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
			case MM_CLEAR: 
				if(MODEFLAG(MM_REF, MM_FULLDELETE)) {
					ObjForCore(rel_owner)->inDeleteObject();
				}
				else { 
					rel_owner[ATTRID_REFERENCE] = NULLCOREOBJ;
					CoreObjMark(self, OBJEVENT_REFRELEASED);
					CoreObjMark(rel_owner, OBJEVENT_RELATION);
				}
			}
		}
	}
	}
	
	if(n == DTID_MODEL) {
		CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
		ITERATE_THROUGH(children) {
			ObjTreeCheckINTORelations(mgaproject, ITER, internals);
		}
	}
}

void ObjTreeCheckINTORelationsFoldersToo(CMgaProject *mgaproject, CoreObj &self, coreobjhash &internals) {
	metaid_type n = GetMetaID(self);
	ASSERT(n >= DTID_MODEL && n <= DTID_FOLDER);
	if ( n >= DTID_MODEL && n <= DTID_SET)
	{
		CoreObjs xrefs = self[ATTRID_XREF + ATTRID_COLLECTION]; 
		ITERATE_THROUGH(xrefs) {
			metaid_type st = GetMetaID(ITER);
			if(st == DTID_SETNODE || st == DTID_CONNROLE) {
				CoreObj rel_owner = ITER.GetMgaObj();
				if(internals.find(rel_owner) == internals.end()) {
					int ttt = st == DTID_CONNROLE ? MM_CONN : MM_SET;
					setcheck(mgaproject, rel_owner, CHK_CHANGED);
					switch(MODEMASK(ttt, MM_INTO)) {
					case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
						break;
					case MM_CLEAR: 
						if( st == DTID_CONNROLE && ObjForCore(rel_owner)->simpleconn() ||
							MODEFLAG(ttt, MM_FULLDELETE)) {
							ObjForCore(rel_owner)->inDeleteObject();
						} 
						else {
							CoreObjMark(self, st == DTID_CONNROLE ? OBJEVENT_DISCONNECTED : OBJEVENT_SETEXCLUDED);
							CoreObjMark(rel_owner, OBJEVENT_RELATION);
							SingleObjTreeDelete(ITER);
						}
						break;
					}
				}	
			}
		}
		{
			CoreObjs refs = self[ATTRID_REFERENCE + ATTRID_COLLECTION]; 
			ITERATE_THROUGH(refs) {
				CoreObj rel_owner = ITER;
				if(internals.find(rel_owner) == internals.end()) {
					setcheck(mgaproject, rel_owner, CHK_CHANGED);
					switch(MODEMASK(MM_REF, MM_INTO)) {
					case MM_ERROR: COMTHROW(E_MGA_OP_REFUSED);
					case MM_CLEAR: 
						if(MODEFLAG(MM_REF, MM_FULLDELETE)) {
							ObjForCore(rel_owner)->inDeleteObject();
						}
						else { 
							rel_owner[ATTRID_REFERENCE] = NULLCOREOBJ;
							CoreObjMark(self, OBJEVENT_REFRELEASED);
							CoreObjMark(rel_owner, OBJEVENT_RELATION);
						}
					}
				}
			}
		}
	}
	if(n == DTID_MODEL || n == DTID_FOLDER) {
		CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
		ITERATE_THROUGH(children) {
			ObjTreeCheckINTORelationsFoldersToo(mgaproject, ITER, internals);
		}
	}
}

void ObjTreeNotify(CMgaProject *mgaproject, CoreObj &self) {
	int k = OBJEVENT_CREATED;
	metaid_type typ = GetMetaID(self);
	if(typ == DTID_FOLDER || typ == DTID_MODEL) {
			CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
			ITERATE_THROUGH(children) {
				k |= OBJEVENT_NEWCHILD;
				ObjTreeNotify(mgaproject, ITER);
			}
	}
	else if(typ == DTID_REFERENCE) {
			CoreObj rr = self[ATTRID_REFERENCE];
			if(rr) { 
				CComPtr<IMgaFCO> fco;
				ObjForCore(rr)->getinterface(&fco);
				mgaproject->ObjMark(fco, OBJEVENT_REFERENCED);
			}
	}
	else if(typ == DTID_CONNECTION || typ == DTID_SET) {
			
			CoreObjs rrs;
			long msg;
			if(typ == DTID_SET) {
				rrs = self[ATTRID_SETMEMBER+ATTRID_COLLECTION];
				msg = OBJEVENT_SETINCLUDED;
			}
			else {
				rrs = self[ATTRID_CONNROLE+ATTRID_COLLECTION];
				msg = OBJEVENT_CONNECTED;
			}
			ITERATE_THROUGH(rrs) { 
				CoreObj fff = ITER[ATTRID_XREF];
				if(fff) {
					CComPtr<IMgaFCO> fco;
					ObjForCore(fff)->getinterface(&fco);
					mgaproject->ObjMark(fco, msg);
				}
			}
	}
	ObjForCore(self)->SelfMark(k);
}

// check of obj or any of its children is derived from root
void CheckConflict(CoreObj &b, CoreObj &root) {

	if(GetMetaID(b) == DTID_MODEL) {
		CoreObj base = b[ATTRID_DERIVED];
		while(base) {
			if(COM_EQUAL(base, root)) COMTHROW(E_MGA_OP_REFUSED);
			base = base[ATTRID_DERIVED];
		}
		CoreObjs children = b[ATTRID_FCOPARENT + ATTRID_COLLECTION];
		ITERATE_THROUGH(children) {
			CheckConflict(ITER, root);
		}
	}		
}


/////////////////
// DeriveNewObjs
/////////////////
///////////////////
// !!! RECURSIVE 
///////////////////
void ReDeriveNewObjs(CMgaProject *mgaproject, std::vector<CoreObj> &orignobjs, int cnt, int targetlevel) {
	ASSERT(targetlevel >= 1);
	if(cnt < 1) return;
	CoreObj orig = orignobjs[0][ATTRID_PARENT];
	if(!orig.IsFCO()) return;
	CoreObjs deriveds = orig[ATTRID_DERIVED+ATTRID_COLLECTION];
	ITERATE_THROUGH(deriveds) {
		int i;
		CoreObj subtypebase = ITER.FollowChain(ATTRID_PARENT, targetlevel);
		coreobjpairhash derivcrealist;
		coreobjhash derivmovedlist;
		long instance = ITER[ATTRID_PERMISSIONS] & INSTANCE_FLAG;
		std::vector<CoreObj> newderiveds(cnt);
		for(i = 0; i< cnt; i++) {
			ObjTreeDerive(mgaproject, orignobjs[i], newderiveds[i], derivcrealist, instance);
			newderiveds[i][ATTRID_FCOPARENT]=ITER;
		}
		if(cnt) CoreObjMark(ITER, OBJEVENT_NEWCHILD);
// Reroute references
		for(i = 0; i< cnt; i++) {
			ObjTreeReconnect(newderiveds[i], derivcrealist, subtypebase);
		}
		shiftlist(derivcrealist, derivmovedlist);
		for(i = 0; i< cnt; i++) {
			ObjTreeCheckRelations(mgaproject, newderiveds[i], derivmovedlist);			
		}
		ReDeriveNewObjs(mgaproject, newderiveds, cnt, targetlevel);
	}
}


void ReDeriveNewObj(CMgaProject *mgaproject, CoreObj &orignobj, int targetlevel) {
   std::vector<CoreObj> vv(1);
   vv[0] = orignobj;
   ReDeriveNewObjs(mgaproject, vv, 1, targetlevel);
}



/////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// COPY //////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////


//##ModelId=3C5AA0970352
HRESULT FCO::CopyFCOs(IMgaFCOs *copylist, IMgaMetaRoles *rlist,IMgaFCOs **objs) {
	COMTRY_IN_TRANSACTION {
		CheckWrite();
		CHECK_MYINPTRSPAR(copylist);
		long cnt;
		COMTHROW(copylist->get_Count(&cnt));
		if(rlist) {
			long rlcnt;
			COMTHROW(rlist->get_Count(&rlcnt));
			if(rlcnt != cnt) COMTHROW(E_MGA_BAD_COLLENGTH);
		}
		metaid_type targettype = GetMetaID(self);
		int targetlevel = 0;
		CoreObj rootp;
// Pre check:
		if(targettype == DTID_MODEL) {
			if(self[ATTRID_PERMISSIONS] & ~EXEMPT_FLAG) COMTHROW(E_MGA_NOT_CHANGEABLE);
			GetRootOfDeriv(self, rootp, &targetlevel);
			if(rootp) {
				MGACOLL_ITERATE(IMgaFCO, copylist) {
					CheckConflict(CoreObj(MGACOLL_ITER), rootp);
				} MGACOLL_ITERATE_END;
			}
		}

// Copy trees
		coreobjpairhash crealist;
		int i = 0;

		std::vector<CoreObj> nobjs(cnt);
		MGACOLL_ITERATE(IMgaFCO, copylist) {
			CoreObj oldobj = CoreObj(MGACOLL_ITER);
			int derdist = GetRealSubtypeDist(oldobj);
			ObjTreeCopy(mgaproject, oldobj, nobjs[i], crealist);  // copy
			if(derdist) ObjTreeDist(nobjs[i], derdist);
			assignnewchild(nobjs[i]);
			metaref_type trole = METAREF_NULL;
			if(targettype != DTID_FOLDER) {
				ASSERT(targettype == DTID_MODEL);
				CComPtr<IMgaMetaRole> r;
				if(rlist) COMTHROW(rlist->get_Item(i+1, &r));
				if(!r) {   // NO metarole given, inherit that of original object
					CComPtr<IMgaMetaFCO> mf; 
					COMTHROW(get_Meta(&mf));
					CComQIPtr<IMgaMetaModel> parentmeta = mf;
					if(!parentmeta) COMTHROW(E_MGA_META_INCOMPATIBILITY);
					metaref_type t;
					t = (nobjs[i])[ATTRID_ROLEMETA];
					if(!t) COMTHROW(E_MGA_NO_ROLE);
					CComQIPtr<IMgaMetaRole> metar = mgaproject->FindMetaRef(t);
					if(!metar) COMTHROW(E_MGA_META_INCOMPATIBILITY);
					CComBSTR rolename;
					COMTHROW(metar->get_Name(&rolename));
					COMTHROW(parentmeta->get_RoleByName(rolename, &r));
					if(!r) COMTHROW(E_MGA_NO_ROLE);

				}
				{
					metaref_type kt;
					CComPtr<IMgaMetaFCO> mfco;
					COMTHROW(r->get_Kind(&mfco));
					COMTHROW(mfco->get_MetaRef(&kt));
					if(kt != (nobjs[i])[ATTRID_META]) COMTHROW(E_MGA_NO_ROLE);
				}

				COMTHROW(r->get_MetaRef(&trole));
				if(trole == METAREF_NULL)  COMTHROW(E_MGA_INVALID_ROLE);
			}
			(nobjs[i])[ATTRID_ROLEMETA] = trole;
			i++;
		} MGACOLL_ITERATE_END;

// Reroute references
		for(i = 0; i< cnt; i++) {
			ObjTreeReconnect(nobjs[i], crealist);			
		}

		coreobjhash newcrealist;
		shiftlist(crealist, newcrealist);

		for(i = 0; i< cnt; i++) {
			ObjTreeCheckRelations(mgaproject, nobjs[i], newcrealist);			
		}

		if(targetlevel >= 0) ReDeriveNewObjs(mgaproject, nobjs, cnt, targetlevel+1);

		docheck(mgaproject);

// Assemble return array:
		if(objs) {
			CREATEEXCOLLECTION_FOR(MgaFCO, q);
			for(i = 0; i< cnt; i++) {
				CComPtr<IMgaFCO> ff;
				ObjForCore(nobjs[i])->getinterface(&ff);
				q->Add(ff); 
			}
			*objs = q.Detach();
		}



		SelfMark(OBJEVENT_NEWCHILD);

	} COMCATCH_IN_TRANSACTION(;);
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// MOVE //////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////


// give new value to subobj, and also give the same value to all its slave subobjs
void RedirectMasterTree(CoreObj &subobj, attrid_type ai, CoreObj &newval) {
	subobj[ai] = newval;
	CoreObjs slso;
	slso = subobj[ATTRID_MASTEROBJ+ATTRID_COLLECTION];
	ITERATE_THROUGH(slso) {
		RedirectMasterTree(ITER, ai, newval);
	}
}

// An object with tree has been moved into a type, fixup relations that used to point from inside
// to the object just moved, i.e. which have just became 'internal'
// self = the target object that may have too many pointers pointing at itself
// selfbase = the base of the subtype/instance (self is contained by that)
// level = the distance between self and selfbase

void ObjTreeInternalize(CMgaProject *mgaproject, CoreObj &self, CoreObj &selfbase, int tlevel) {
	CoreObj origbase = selfbase[ATTRID_DERIVED];
	CoreObj orig = self[ATTRID_DERIVED];
	static attrid_type ais[] = {ATTRID_SEGREF, ATTRID_REFERENCE, ATTRID_XREF };
// do loop for ATTRID_REFERENCE, ATTRID_XREF, and also for ATTRID_SEGREF in case of ref objs
	for(int i = (self.GetMetaID() == DTID_REFERENCE) ? 0 : 1; i < DIM(ais); i++) {
		attrid_type ai = ais[i];
		CoreObjs rels = orig[ai + ATTRID_COLLECTION];
		ITERATE_THROUGH(rels) {
			metaid_type typ = ITER.GetMetaID();
// 1. The object must be CONNROLE or SETNODE
			if(typ != DTID_CONNROLE && typ != DTID_SETNODE) continue;
// 2. The object must not have a master pointing to the same object
			CoreObj obmaster = ITER[ATTRID_MASTEROBJ];
			if(obmaster && COM_EQUAL(orig, CoreObj(obmaster[ai]))) continue;
			CoreObj srcobj = ITER.GetMgaObj();
			int srclevel;
// 3. The relation object must be part of the same type (i.e common derivation parent)
			if(!IsContained(srcobj, origbase, &srclevel)) continue;	 // it is an external reference

			CoreObjs slso = ITER[ATTRID_MASTEROBJ+ATTRID_COLLECTION];
	
			ITERATE_THROUGH(slso) {
				CoreObj derivedsrc = ITER.GetMgaObj();
				if(COM_EQUAL(selfbase, derivedsrc.FollowChain(ATTRID_FCOPARENT, srclevel))) {
					ASSERT(COM_EQUAL(orig, CoreObj(ITER[ai])));
					RedirectMasterTree(ITER,ai,self);
					setcheck(mgaproject, derivedsrc, CHK_CHANGED);
					break;
				}
			}
		}
	}
// descend to children, increase level
	if(self.GetMetaID() == DTID_MODEL) {
		CoreObjs children = self[ATTRID_FCOPARENT + ATTRID_COLLECTION];
		ITERATE_THROUGH(children) {
			ObjTreeInternalize(mgaproject, ITER, selfbase, tlevel+1);
		}
	}
}




/////////////////
// DeriveMoveds
/////////////////
//   We have set of objects ('orignobjs') moved into a single model in a (base)type. 
//   Reflect changes in its deriveds. 
//   -- 'extmoved' objects must be 
//		- newly derived,  
//		- 'internalized'
//		- reconnected
//   -- others must be located and moved here
//   They all need to be checked.
//   
///////////////////
// !!! RECURSIVE 
///////////////////
void DeriveMoveds(CMgaProject *mgaproject, std::vector<CoreObj> &orignobjs, std::vector<int> &extmoved, int cnt, int targetlevel)
{
	if(cnt < 1) return;

	CoreObj orig = orignobjs[0][ATTRID_FCOPARENT];
	metaid_type n = GetMetaID( orig);

	if (n >= DTID_MODEL && n <= DTID_SET)
	{
		CoreObjs deriveds = orig[ATTRID_DERIVED+ATTRID_COLLECTION];
		ITERATE_THROUGH(deriveds) {
			int i;
			CoreObj subtypebase = ITER.FollowChain(ATTRID_FCOPARENT, targetlevel);
			coreobjpairhash derivcrealist;
			coreobjhash derivmovedlist;
			long instance = ITER[ATTRID_PERMISSIONS] & INSTANCE_FLAG;
			std::vector<CoreObj> newderiveds(cnt);
			for(i = 0; i< cnt; i++) {
				if(extmoved[i] < 0) {	// the movedobject is a newcomer in the basetype, derive new obj in subtype
					ObjTreeDerive(mgaproject, orignobjs[i], newderiveds[i], derivcrealist, instance);
					newderiveds[i][ATTRID_FCOPARENT]=ITER;
					ObjTreeInternalize(mgaproject, newderiveds[i], subtypebase, targetlevel+1);
				}
				else {
					CoreObj tgt(ITER);	// locate the corresponding object in subtype and move it
					CoreObjs ders = orignobjs[i][ATTRID_DERIVED+ATTRID_COLLECTION];
					ITERATE_THROUGH(ders) {
						CoreObj der(ITER);
						if(COM_EQUAL(subtypebase, ITER.FollowChain(ATTRID_FCOPARENT, extmoved[i]))) {
								newderiveds[i] = ITER;
								CoreObjMark(ITER[ATTRID_FCOPARENT], OBJEVENT_LOSTCHILD);
								ITER[ATTRID_FCOPARENT] = tgt;
								ITER[ATTRID_ROLEMETA] = orignobjs[i][ATTRID_ROLEMETA];
								ObjTreeCollect(mgaproject, ITER, derivmovedlist, CHK_MOVED);
								break;
						}
					}
					ASSERT(ITER_BROKEN);
				}
			}
			CoreObjMark(ITER, OBJEVENT_NEWCHILD);
	// Reroute references
			for(i = 0; i< cnt; i++) {
				ObjTreeReconnect(newderiveds[i], derivcrealist, subtypebase);
			}
			shiftlist(derivcrealist, derivmovedlist);
			for(i = 0; i< cnt; i++) {
				ObjTreeCheckRelations(mgaproject, newderiveds[i], derivmovedlist);			
			}
			DeriveMoveds(mgaproject, newderiveds, extmoved, cnt, targetlevel);

			for(i = 0; i< cnt; i++) {
				ObjTreeCheckINTORelations(mgaproject, newderiveds[i], derivmovedlist);			
			}
		}
	}
}


//##ModelId=3C5AA097037A
HRESULT FCO::MoveFCOs(IMgaFCOs *movelist, IMgaMetaRoles *rlist,IMgaFCOs **objs) {
	COMTRY_IN_TRANSACTION {
		CheckWrite();
		CHECK_MYINPTRSPAR(movelist);
		long cnt;
		COMTHROW(movelist->get_Count(&cnt));

		bool valid = ( cnt > 0);
		if ( cnt == 1)
		{
			valid = false;
			// check whether the source folder is the target as well
			CComPtr<IMgaFCO> mf;
			MGACOLL_ITERATE( IMgaFCO, movelist) {
				mf = MGACOLL_ITER;
			} MGACOLL_ITERATE_END;

			VARIANT_BOOL is_equal;
			IMgaFCO * thisptr;
			getinterface( &thisptr);
			COMTHROW( mf->get_IsEqual( thisptr, &is_equal));

			if (is_equal == VARIANT_FALSE) // not equal
				//COMTHROW( E_MGA_INVALID_ARG); // do not copy/move onto itself
				valid = true;
		}

		if ( valid) 
		{
			if(rlist) {
				long rlcnt;
				COMTHROW(rlist->get_Count(&rlcnt));
				if(rlcnt != cnt) COMTHROW(E_MGA_BAD_COLLENGTH);
			}
			metaid_type targettype = GetMetaID(self);
			int targetlevel = 0;
			CoreObj rootp;
// Pre check:
			if(targettype == DTID_MODEL) {
				if(self[ATTRID_PERMISSIONS]  & ~EXEMPT_FLAG) COMTHROW(E_MGA_NOT_CHANGEABLE);
				GetRootOfDeriv(self, rootp, &targetlevel);
				MGACOLL_ITERATE(IMgaFCO, movelist) {
					CoreObj cur = CoreObj(MGACOLL_ITER);
					if(IsContained(self, cur)) COMTHROW(E_MGA_OP_REFUSED);
					if(cur[ATTRID_RELID] >= RELIDSPACE) COMTHROW(E_MGA_OP_REFUSED); 
					if(rootp) {
						CheckConflict(cur, rootp);
					}
				} MGACOLL_ITERATE_END;
			}

// move trees
			coreobjhash moveslist;
			int i = 0;

			std::vector<CoreObj> nobjs(cnt);
			std::vector<int> moved_into(cnt);  // >= 0 for objects moved inside the tree, -1 for newcomers
			MGACOLL_ITERATE(IMgaFCO, movelist) {
				CoreObj cur = nobjs[i] = CoreObj(MGACOLL_ITER);
				CoreObj movedobjrootp;
				int movedobjlevel;
				GetRootOfDeriv(cur, movedobjrootp, &movedobjlevel);
				int derdist = GetRealSubtypeDist(cur);
				if(derdist) COMTHROW(E_MGA_OP_REFUSED);
	//			if(derdist)	ObjTreeDist(cur, derdist);  // move

				CoreObj curp = cur[ATTRID_FCOPARENT];
				if(!COM_EQUAL(curp, self)) {
					CoreObjMark(curp, OBJEVENT_LOSTCHILD);
					assignnewchild(cur);
					CoreObjMark(cur, OBJEVENT_PARENT);
				}
				ObjTreeCollect(mgaproject, cur, moveslist, CHK_MOVED);
				metaref_type trole = METAREF_NULL;
				if(targettype != DTID_FOLDER) {
					ASSERT(targettype == DTID_MODEL);
					CComPtr<IMgaMetaRole> r;
					if(rlist) COMTHROW(rlist->get_Item(i+1, &r));
					if(!r) {   // NO metanaem given, inherit that of original object
						CComPtr<IMgaMetaFCO> mf; 
						COMTHROW(get_Meta(&mf));
						CComQIPtr<IMgaMetaModel> parentmeta = mf;
						if(!parentmeta) COMTHROW(E_MGA_META_INCOMPATIBILITY);
						cur[ATTRID_ROLEMETA];
						metaref_type t;
						t = (nobjs[i])[ATTRID_ROLEMETA];
						if(!t) COMTHROW(E_MGA_NO_ROLE);
						CComQIPtr<IMgaMetaRole> metar = mgaproject->FindMetaRef(t);
						if(!metar) COMTHROW(E_MGA_META_INCOMPATIBILITY);
						CComBSTR rolename;
						COMTHROW(metar->get_Name(&rolename));
						COMTHROW(parentmeta->get_RoleByName(rolename, &r));
						if(!r) COMTHROW(E_MGA_NO_ROLE);
					}
					{
						metaref_type kt;
						CComPtr<IMgaMetaFCO> mfco;
						COMTHROW(r->get_Kind(&mfco));
						COMTHROW(mfco->get_MetaRef(&kt));
						if(kt != cur[ATTRID_META]) COMTHROW(E_MGA_META_INCOMPATIBILITY);
					}

					COMTHROW(r->get_MetaRef(&trole));
					if(trole == METAREF_NULL)  COMTHROW(E_MGA_INVALID_ROLE);
				}
				cur[ATTRID_ROLEMETA] = trole;

				// Determine what to do with subtypes of the objects moved 
				CoreObjs ders = cur[ATTRID_DERIVED+ATTRID_COLLECTION];
				if(!rootp || !COM_EQUAL(movedobjrootp, rootp)) {     // ********* moving out of a type
					moved_into[i] = -1;				   // ******** true if moving into a type
					if(movedobjlevel < 1) {	// not derived, or derived as self
											// if the object to be moved is a root of derivs,
											// and not moved to a model, thst model must not have any subtypes/instances
						if(ders.Count() &&
							targettype == DTID_MODEL && 
							CoreObjs(self[ATTRID_DERIVED+ATTRID_COLLECTION]).Count()) {
								COMTHROW(E_MGA_OP_REFUSED);  
						}
					}
					else {
						ITERATE_THROUGH(ders) {				   // ******* erase all previous subtypes
							ObjForCore(ITER)->inDeleteObject();
						}
					}
				}
				else moved_into[i] = movedobjlevel;        // ***** the object remains under the same rootmodel
				i++;
			} MGACOLL_ITERATE_END;

			for(i = 0; i< cnt; i++) {
				ObjTreeCheckRelations(mgaproject, nobjs[i], moveslist);			
			}

			if(targettype == DTID_MODEL) {
				DeriveMoveds(mgaproject, nobjs, moved_into, cnt, targetlevel);
			}

// Reroute references
			for(i = 0; i< cnt; i++) {
				ObjTreeCheckINTORelations(mgaproject, nobjs[i], moveslist);			
			}


			docheck(mgaproject);

// Assemble return array:
			CREATEEXCOLLECTION_FOR(MgaFCO, q);
			for(i = 0; i< cnt; i++) {
				CComPtr<IMgaFCO> n;
				ObjForCore(nobjs[i])->getinterface(&n);
				q->Add(n); 
			}

			
			if(objs) {
				*objs = q.Detach();
			}


			SelfMark(OBJEVENT_NEWCHILD);
		} // if valid

	} COMCATCH_IN_TRANSACTION(;);
}




/////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////// DERIVE //////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////

// do not use with folders
bool IsAnythingDerivedFromChildren(CoreObj cbase) {
	if(GetMetaID(cbase) == DTID_MODEL) {
		ITERATE_THROUGH(CoreObjs(cbase[ATTRID_FCOPARENT + ATTRID_COLLECTION])) {
			if(CoreObjs(ITER[ATTRID_DERIVED + ATTRID_COLLECTION]).Count() ||
				IsAnythingDerivedFromChildren(ITER) ) return true;
		}
	}
	return false;
}



//##ModelId=3C5AA09703A2
HRESULT FCO::DeriveFCO(IMgaFCO *base, IMgaMetaRole *role, VARIANT_BOOL binstance, IMgaFCO** newobj) {
	COMTRY_IN_TRANSACTION {
		CheckWrite();
		CHECK_INBOOLPAR(binstance);
		CHECK_MYINPTRPAR(base);
		CHECK_OUTPTRPAR(newobj);
		metaid_type targettype = GetMetaID(self);

// Pre check:
// ERROR if object is not rootobject
		objtype_enum l;
		CoreObj cbase(base);
		COMTHROW(base->get_ObjType(&l));
		if(l == OBJTYPE_FOLDER) COMTHROW(E_MGA_NOT_DERIVABLE);
//NCH
//		COMTHROW(base->GetParent(NULL, &l));
//		if(l != OBJTYPE_FOLDER) COMTHROW(E_MGA_NOT_DERIVABLE);

//		Derivable: the FCO parent of base has 0 deriveds
//				   base already has deriveds, or all the children have 0 deriveds too.
		CoreObj parent = cbase[ATTRID_FCOPARENT];
		int lev;
		CoreObj rr;
		GetRootOfDeriv(cbase, rr, &lev);
		if( lev > 0 ||
			lev == -1 && IsAnythingDerivedFromChildren(cbase) ) {
				COMTHROW(E_MGA_NOT_DERIVABLE);
		}



// ERROR If new base is already an instance
		VARIANT_BOOL parentinst;
		COMTHROW(base->get_IsInstance(&parentinst));
		if(parentinst) COMTHROW(E_MGA_NOT_DERIVABLE);
// Check if object to be created or any of its children will become children of their own basetypes
		int targetlevel = 0;
		if(targettype != DTID_FOLDER) {
			if(self[ATTRID_PERMISSIONS] & ~EXEMPT_FLAG) COMTHROW(E_MGA_NOT_CHANGEABLE);
			CoreObj root;
			GetRootOfDeriv(self, root, &targetlevel);
			if(root) {
				CoreObj c1(base);
				if(COM_EQUAL(root, c1)) COMTHROW(E_MGA_NOT_DERIVABLE);
				CheckConflict(c1, root);
			}
		}


		
// Copy trees
		CoreObj newcoreobj;
		coreobjpairhash crealist;
		ObjTreeDerive(mgaproject, cbase, newcoreobj, crealist, binstance ? INSTANCE_FLAG : 0);  // copy
		assignnewchild(newcoreobj);

		if(targettype == DTID_FOLDER) newcoreobj[ATTRID_ROLEMETA] = METAREF_NULL;

		else {
			metaref_type t;
			if(!role)  COMTHROW(E_MGA_INVALID_ROLE);
			COMTHROW(role->get_MetaRef(&t));
			if(t == METAREF_NULL)  COMTHROW(E_MGA_INVALID_ROLE);
			{
				metaref_type kt;
				CComPtr<IMgaMetaFCO> mfco;
				COMTHROW(role->get_Kind(&mfco));
				COMTHROW(mfco->get_MetaRef(&kt));
				if(kt != newcoreobj[ATTRID_META]) COMTHROW(E_MGA_META_INCOMPATIBILITY);
			}

			newcoreobj[ATTRID_ROLEMETA] = t;
		}


// Reroute references
		ObjTreeReconnect(newcoreobj, crealist);			
// Post check:
		coreobjhash newcrealist;
		shiftlist(crealist, newcrealist);

		ObjTreeCheckRelations(mgaproject, newcoreobj, newcrealist);			

		ObjTreeNotify(mgaproject, newcoreobj);

		if(targetlevel >= 0) ReDeriveNewObj(mgaproject, newcoreobj, targetlevel+1);

		SelfMark(OBJEVENT_NEWCHILD);
		ObjFor(base)->SelfMark(OBJEVENT_SUBT_INST);

		docheck(mgaproject);


// Notification
		ObjForCore(newcoreobj)->getinterface(newobj);
	} COMCATCH_IN_TRANSACTION(;)
}




