#include "stdafx.h"

#include "MgaClosure.h"
#include "ClosureDlg.h"
#include "ClosureRes.h"
#include "ClosureSmart.h"
#include "MakeClosure.h"


STDMETHODIMP CMgaClosure::SelectiveClosure(
	IMgaFCOs *in_fcos, IMgaFolders *in_folders, 
	IMgaFCOs **sel_fcos, IMgaFolders **sel_folders, 
	IMgaFCOs **top_fcos, IMgaFolders **top_folders, 
	int fc,	
	long *pOptions,
	BSTR *pAcceptingKinds
	) 
{
	AFX_MANAGE_STATE( AfxGetStaticModuleState());
	ASSERT( pOptions);

	CWaitCursor w;
	CClosureDlg dlg;
	dlg.m_activeFolderContainment = fc != 0;
	if ( dlg.DoModal() == IDOK)
	{
		int kind_map = 0;//important to set it to 0
		if( dlg.m_includeAllKindsRadio == 0) 
			kind_map = -1;//0xffffffff;
		else
			kind_map = kind_map
				| (dlg.m_includeFolders?0x1:0)
				| (dlg.m_includeModels?0x2:0)
				| (dlg.m_includeAtoms?0x4:0)
				| (dlg.m_includeSets?0x8:0)
				| (dlg.m_includeReferences?0x10:0);

		CMakeClosure mc( 
			in_fcos, in_folders,
			dlg.m_bRefersTo, dlg.m_bReferredBy,
			dlg.m_bConnection, dlg.m_bRefPortConnection,
			dlg.m_bSetMember, dlg.m_bMemberOfSets, 
			// compare the containment options with the direction option
			dlg.m_direction?dlg.m_bContainment:0, !dlg.m_direction?dlg.m_bPartOfModels:0,
			dlg.m_direction?dlg.m_bFolderContainment:0, !dlg.m_direction?dlg.m_bPartOfFolders:0, 
			dlg.m_bBaseTypes, dlg.m_bDerivedTypes, 
			//dlg.m_bDigIntoLibraries, 
			dlg.m_libraryHandling /*== 2*/, // in case of "continue" ("go on") only
			dlg.m_wrappingOption,
			dlg.m_direction,
			kind_map);

		mc.process(); // do the hard job

		// load the results
		mc.getFCOs( sel_fcos);
		mc.getFolders( sel_folders);

		bool normal_tops = !mc.noTopObjs();
		
		if ( normal_tops) // the top fcos/folds is calculated and The RootFolder is not among them
		{
			mc.getTopFCOs( top_fcos);
			mc.getTopFolders( top_folders);
		}

		
		if( normal_tops &&					// avoid in case of non-normal tops
			dlg.m_wrappingOption > 0 &&		// avoid in case of miniproject
			mc.isAnyAccepting())			// avoid if is no accepting kind/folder
		{
			std::string res_str;
			//CClosureRes res_dlg;

			//res_dlg.m_metaKindsAndFolders = mc.getAccKindsVector();
			mc.getAccKindsInString( res_str);

			// shows the acc kinds in a dialog
			//res_dlg.DoModal(); 
			// shows the acc kinds in the console window
			//mc.send2Console( "[Closure] The following kinds may accept the closure:" + res_str + ".");

			CComBSTR acc_kind = res_str.c_str();
			*pAcceptingKinds = acc_kind.Detach();
		}

		*pOptions = 0;
		*pOptions = (dlg.m_libraryHandling == 0)?0x1:0; // in case of "stubs" set the options to 1

		if ( dlg.m_outputOption == 1) // dump to a file
		{
			CComPtr<IMgaDumper> dumper;
			COMTHROW( dumper.CoCreateInstance(L"Mga.MgaDumper") );

			switch( dlg.m_wrappingOption) {
			case 0:
				{
					COMTHROW( dumper->DumpClos( *sel_fcos, *sel_folders, PutInBstr( dlg.m_strOutputFile), *pOptions) );

					break;
				}
			case 1:
			case 2:
				{
					if ( normal_tops)
						COMTHROW( dumper->DumpClosR( *sel_fcos, *sel_folders, PutInBstr( dlg.m_strOutputFile), *top_fcos, *top_folders, *pOptions, 0, *pAcceptingKinds) );
					else 
						// dump like in miniproject case from The RootFolder
						// if the data is copied to clipboard then the DumpClosR will call the 
						// DumpClos when it will realize the topfco and topfolders are empty
						COMTHROW( dumper->DumpClos( *sel_fcos, *sel_folders, PutInBstr( dlg.m_strOutputFile), *pOptions) );

					*top_fcos = 0;
					*top_folders = 0;
					break;
				}
			} // switch

			*sel_fcos = 0;
			*sel_folders = 0;
		}
		// else: clipboard
	}
	return S_OK;
}

STDMETHODIMP CMgaClosure::SmartCopy(
	IMgaFCOs *in_fcos, IMgaFolders *in_folders,
	IMgaFCOs **sel_fcos, IMgaFolders **sel_folders,
	IMgaFCOs **top_fcos, IMgaFolders **top_folders,
	int options_in, 
	long *pOptions,
	BSTR *accept,
	BSTR *path
	)
{
	AFX_MANAGE_STATE( AfxGetStaticModuleState());
	ASSERT( pOptions);

	bool res;
	CWaitCursor w;
	CClosureSmart dlg;
	CKindSelectClosure *mc;

	bool any_container = (options_in & 1024) == 1024; // container (model, folder)
	if( any_container)
	{
		dlg.m_disable = false;

		res = dlg.DoModal() == IDOK;
		mc = new CContainerSelectSomeKindClosure
			( in_fcos
			, in_folders
			, dlg.m_bRefs
			, dlg.m_bConns
			, dlg.m_bSets
			);
	}
	else
	{
		dlg.m_disable = true;
		dlg.m_bRefs = (options_in & 2048) == 2048; // is any ref
		dlg.m_bSets = (options_in & 4096) == 4096; // is any set 
		dlg.m_bConns = (options_in & 8192) == 8192; // is any conn

		res = dlg.DoModal() == IDOK;

		mc = new CKindSelectClosure
			( in_fcos
			, in_folders
			, dlg.m_bRefs
			, dlg.m_bConns
			, dlg.m_bSets
			);
	}

	if( res)
	{
		mc->process(); // do the hard job

		// load the results
		mc->getFCOs( sel_fcos);
		mc->getFolders( sel_folders);

		bool normal_tops = !mc->noTopObjs();
		
		if( normal_tops) // the top fcos/folds is calculated and The RootFolder is not among them
		{
			mc->getTopFCOs( top_fcos);
			mc->getTopFolders( top_folders);
		}

		*pOptions = 0;
		//*pOptions = (dlg.m_libraryHandling == 0)?0x1:0; // in case of "stubs" set the options to 1
		*pOptions = (LIB_STUBS == 0)?0x1:0; // in case of "stubs" set the options to 1
		*pOptions = *pOptions | 0x2; // dump the closure2 like attributes
		*pOptions = *pOptions | 0x4; // smart copy signal

		CComBSTR acckind = mc->m_accKind.c_str();
		*accept = acckind.Detach();

		CComBSTR name = mc->m_iniSelPath.c_str();
		*path = name.Detach();
	}

	delete mc;

	return S_OK;
}
